﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright (c) 2023 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "stm32f3xx.h"
#include "stm32f3xx_ll_rcc.h"
#include "stm32f3xx_ll_bus.h"
#include "stm32f3xx_ll_system.h"
#include "stm32f3xx_ll_utils.h"
#include "stm32f3xx_ll_pwr.h"
#include "stm32f3xx_ll_usart.h"
#include "stm32f3xx_ll_gpio.h"
#include "ARMChip.h"
#include "Conf.h"
#include "Board.h"
#include "Flash.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt32   		launchCode  __attribute__ ((section (".noinit")));
static volatile UInt32  msecTicks = 0;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt32
MSecTicks(void)
{
	UInt32 ticks;
	Atomic
	{	ticks = msecTicks;
	}
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Vorlauf  /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((constructor)) void
PreStart(void)
{
	bool powerOn        = (bool)LL_RCC_IsActiveFlag_PORRST();
	bool windowWatchdog = (bool)LL_RCC_IsActiveFlag_WWDGRST();
	bool resetPin       = (bool)LL_RCC_IsActiveFlag_PINRST();
	bool software       = (bool)LL_RCC_IsActiveFlag_SFTRST();
	bool watchdog       = (bool)LL_RCC_IsActiveFlag_IWDGRST();
	if (powerOn)
	{	launchCode = ~CFG_PGMLDR_MARKER;
	}
	else if (windowWatchdog)
	{	launchCode = CFG_PGMLDR_MARKER;
	}
	else if (!software && !watchdog && resetPin)
	{	launchCode = CFG_PGMLDR_MARKER;
	}
	if (launchCode != CFG_PGMLDR_MARKER)
	{	// Wenn Anwendung vorhanden ...
		extern const char APPL_SECTION_START[8];
		const UInt32* applStartAddress = (const UInt32*)(&APPL_SECTION_START);
		if (applStartAddress[0] != 0xFFFFFFFF && applStartAddress[1] != 0xFFFFFFFF)
	 	{	// ... Anwendung starten
			const  UInt32 stackAddress = applStartAddress[0];
			const  UInt32 resetHandler = applStartAddress[1];
			asm("msr msp, %0; bx %1;" : : "r"(stackAddress), "r"(resetHandler));
		}
		launchCode = CFG_PGMLDR_MARKER;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
SysTick_Handler(void)
{
	msecTicks++;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Setup(void)
{
    DisableInterrupts();
    LL_RCC_ClearResetFlags();
    // Watchdog --------------------------------------------------------------------------
    {	IWDG->KR  = 0x0000CCCC;					// Watchdog aktivieren
    	IWDG->KR  = 0x00005555; 				// Schreibzugriff auf PR, RLR und WINR
    	IWDG->PR  = IWDG_PR_PR_1| IWDG_PR_PR_2;	// Vorteiler: 40 kHz ÷ 256 = 156 kHz
    	IWDG->RLR = 2*156;						// Watchdog-Zeit auf 2 s
    	while (IWDG->SR);						// Warten, bis Änderungen abschlossen sind
    	IWDG->KR  = 0x0000AAAA;					// WD zurücksetzen/Schreibzugr. verbieten
    }
    // Systemtakte initialisieren --------------------------------------------------------
    {	LL_FLASH_SetLatency(LL_FLASH_LATENCY_0);
  		while(LL_FLASH_GetLatency()!= LL_FLASH_LATENCY_0);
  		LL_RCC_HSE_EnableBypass();
  		LL_RCC_HSE_Enable();
  		while(LL_RCC_HSE_IsReady() != 1);
  		LL_RCC_SetAHBPrescaler(LL_RCC_SYSCLK_DIV_1);
  		LL_RCC_SetAPB1Prescaler(LL_RCC_APB1_DIV_4);
  		LL_RCC_SetAPB2Prescaler(LL_RCC_APB2_DIV_4);
  		LL_RCC_SetSysClkSource(LL_RCC_SYS_CLKSOURCE_HSE);
  		while(LL_RCC_GetSysClkSource() != LL_RCC_SYS_CLKSOURCE_STATUS_HSE);
  		LL_Init1msTick(8000000);
  		LL_SetSystemCoreClock(8000000);
  		LL_RCC_SetUSARTClockSource(LL_RCC_USART1_CLKSOURCE_PCLK1);
	}
    // Busse aktivieren ------------------------------------------------------------------
    {	LL_APB2_GRP1_EnableClock(LL_APB2_GRP1_PERIPH_SYSCFG);
  		LL_APB1_GRP1_EnableClock(LL_APB1_GRP1_PERIPH_PWR);
  		LL_AHB1_GRP1_EnableClock(LL_AHB1_GRP1_PERIPH_GPIOA);
  		LL_AHB1_GRP1_EnableClock(LL_AHB1_GRP1_PERIPH_GPIOB);
		LL_APB2_GRP1_EnableClock(LL_APB2_GRP1_PERIPH_USART1);
	}
    // GPIOs -----------------------------------------------------------------------------
	{	// Onboard-LED: PB3
		LL_GPIO_InitTypeDef setup = { 0 };
		setup.Pin        = LL_GPIO_PIN_3;
		setup.Mode       = LL_GPIO_MODE_OUTPUT;
		setup.Speed      = LL_GPIO_SPEED_FREQ_LOW;
		setup.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
		setup.Pull       = LL_GPIO_PULL_NO;
		LL_GPIO_Init(GPIOB, &setup);
		BOARD_SetLed();
	}
	// Millisekunden-Zeitgeber -----------------------------------------------------------
	{	SystemCoreClockUpdate();
		SysTick_Config(SystemCoreClock/1000);
		NVIC_SetPriority(SysTick_IRQn, 0);
	}
	// USART 1 19200 Baud, 8 Data, 1 Stop, No Parity, Async. -----------------------------
	{	// USART1_TX: PA9
		{	LL_GPIO_InitTypeDef setup = { 0 };
	  		setup.Pin        = LL_GPIO_PIN_9;
	  		setup.Mode       = LL_GPIO_MODE_ALTERNATE;
	  		setup.Speed      = LL_GPIO_SPEED_FREQ_HIGH;
	  		setup.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
	  		setup.Pull       = LL_GPIO_PULL_NO;
	  		setup.Alternate  = LL_GPIO_AF_7;
	  		LL_GPIO_Init(GPIOA, &setup);
		}
	  	// USART1_RX: PA10
	  	{	LL_GPIO_InitTypeDef setup = { 0 };
	  		setup.Pin        = LL_GPIO_PIN_10;
			setup.Mode       = LL_GPIO_MODE_ALTERNATE;
			setup.Speed      = LL_GPIO_SPEED_FREQ_HIGH;
			setup.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
			setup.Pull       = LL_GPIO_PULL_NO;
			setup.Alternate  = LL_GPIO_AF_7;
			LL_GPIO_Init(GPIOA, &setup);
		}
	  	// USART1: Setup
		{	LL_USART_InitTypeDef setup = { 0 };
			setup.BaudRate            = 19200;
			setup.DataWidth           = LL_USART_DATAWIDTH_8B;
			setup.StopBits            = LL_USART_STOPBITS_1;
			setup.Parity              = LL_USART_PARITY_NONE;
			setup.TransferDirection   = LL_USART_DIRECTION_TX_RX;
			setup.HardwareFlowControl = LL_USART_HWCONTROL_NONE;
			setup.OverSampling        = LL_USART_OVERSAMPLING_16;
			LL_USART_Init(USART1, &setup);
			LL_USART_DisableOverrunDetect(USART1);
			LL_USART_DisableDMADeactOnRxErr(USART1);
			LL_USART_ConfigAsyncMode(USART1);
			LL_USART_Enable(USART1);
	  		NVIC_EnableIRQ(USART1_IRQn);
		}
	}
    // Sonstiges -------------------------------------------------------------------------
	// Hinweis zu Interrupt Prioritäten / Prioritätsgruppen / Subprioritäten:
	// Der Reset-Wert für die Aufteilung der Prioritäten in Prioritätsgruppen und Sub-
	// prioritäten (siehe NVIC_GetPriorityGrouping()) ist 0. Das bedeutet: Es gibt 16
	// Prioritätsgruppen aber keine Subprioritäten. Dazu PM0214, S.41:
	//    "Only the group priority determines preemption of interrupt exceptions. When the
	//    processor is executing an interrupt exception handler, another interrupt with the
	//    same group priority as the interrupt being handled does not preempt the handler."
	// Der Reset-Wert für die Interrupt Prioritäten (siehe NVIC_GetPriority(IRQn)) ist
	// ebenfalls 0. Damit haben alle IRQs die gleiche Prioritätsgruppe und unterbrechen
	// sich laut PM0214 gegenseitig nicht.
    EnableInterrupts();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Restart(void)
{
	launchCode = CFG_PGMLDR_MARKER;
	NVIC_SystemReset();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterAppl(void)
{
	launchCode = ~CFG_PGMLDR_MARKER;
	NVIC_SystemReset();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_TransmitByte(UInt08 byte)
{
    while (!LL_USART_IsActiveFlag_TC(USART1));
    LL_USART_TransmitData8(USART1, byte);
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_WaitForByte(UInt08* byte, UInt32 timeout)
{
	UInt32  timer = BOARD_SetTimeout(timeout);
	while (!BOARD_Timeout(timer))
	{	if (LL_USART_IsActiveFlag_RXNE(USART1))
		{	*byte = LL_USART_ReceiveData8(USART1);
			return true;
		}
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Watchdog(void)
{
	IWDG->KR = 0x0000AAAA;
}

//////////////////////////////////////////////////////////////////////////////////////////
UInt32
BOARD_SetTimeout(UInt32 time)
{
	return MSecTicks() + time;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_Timeout(UInt32 timerHandle)
{
	return ((SInt32)(MSecTicks() - timerHandle)) > 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetLed(void)
{
	LL_GPIO_SetOutputPin(GPIOB, LL_GPIO_PIN_3);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_InvLed(void)
{
	if (LL_GPIO_IsOutputPinSet(GPIOB, LL_GPIO_PIN_3)) LL_GPIO_ResetOutputPin(GPIOB, LL_GPIO_PIN_3);
	else                                              BOARD_SetLed();
}

//////////////////////////////////////////////////////////////////////////////////////////
