﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Flash: Zugriff auf Festspeicher            											//
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2022 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "stm32f3xx.h"
#include "TypeDefs.h"
#include "Flash.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
WaitForIdle(void)
{
	while (TestBit(FLASH->SR, FLASH_SR_BSY));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
Unlock(void)
{
	WaitForIdle();
	if (TestBit(FLASH->CR, FLASH_CR_LOCK))
  	{	FLASH->KEYR = FLASH_KEY1;
    	FLASH->KEYR = FLASH_KEY2;
  	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
Lock(void)
{
	WaitForIdle();
	SetBit(FLASH->CR, FLASH_CR_LOCK);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//	FLASH_PageBaseAddress(): Liefert die Startadresse der Flashspeicherseite Nummer	    //
//	-----------------------  <pageNr>.						                            //
//																						//
//  Eingabe: · pageNr: Seitennummer: 0 .. 'FLASH_PAGE_COUNT'-1.							//
//  Ausgabe: Startadresse der Seite: FLASH_BASE .. FLASH_BASE+FLASH_MEMORY_SIZE-1.  	//
//////////////////////////////////////////////////////////////////////////////////////////
void*
FLASH_PageBaseAddress(PageNr pageNr)
{
	return (void*)(FLASH_BASE + (pageNr.number << FLASH_PAGE_BITS));
}

//////////////////////////////////////////////////////////////////////////////////////////
//	FLASH_PageNumberOf(): Liefert die Nummer der für <address> zuständigen Flashspei-	//
//	--------------------  cherseite.													//
//																						//
//  Eingabe: · address: Speicheradresse: FLASH_BASE .. FLASH_BASE+FLASH_MEMORY_SIZE-1.  //
//  Ausgabe: Seitennummer, in der <address> liegt: 0 .. 'FLASH_PAGE_COUNT'-1.			//																			//
//////////////////////////////////////////////////////////////////////////////////////////
PageNr
FLASH_PageNumberOf(const void* address)
{
	PageNr pageNr = { (((UInt32)address-FLASH_BASE) >> FLASH_PAGE_BITS) };
	return pageNr;
}

//////////////////////////////////////////////////////////////////////////////////////////
//	FLASH_ReadPage(): Liest den Inhalt der Flashspeicherseite Nummer <pageNr> und legt	//
//	----------------  ihn in <data> ab.													//
//																						//
//  Eingabe: · pageNr: Zu lesende Seitennummer: 0 .. 'FLASH_PAGE_COUNT'-1.				//
//           · data:   Bytefeld der Größe 'FLASH_PAGE_SIZE', dass die Seitendaten auf-	//
//                     nimmt.															//
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
void
FLASH_ReadPage(PageNr pageNr, UInt08 data[FLASH_PAGE_SIZE])
{
	const UInt08* flashAddress = FLASH_PageBaseAddress(pageNr);
	WaitForIdle();
	for (UInt32 i=0; i<FLASH_PAGE_SIZE; i++) data[i] = flashAddress[i];
}

//////////////////////////////////////////////////////////////////////////////////////////
//	FLASH_ErasePage(): Löscht die Flashspeicherseite Nummer <pageNr>. Anschließend  	//
//	-----------------  haben alle Bytes der Seite den Wert 0xFF.						//
//																						//
//  Eingabe: · pageNr: Zu löschende Seitennummer: 0 .. 'FLASH_PAGE_COUNT'-1.			//
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
void
FLASH_ErasePage(PageNr pageNr)
{
	Unlock();
	WaitForIdle();
	SetBit(FLASH->CR, FLASH_CR_PER);
    FLASH->AR = (UInt32)FLASH_PageBaseAddress(pageNr);
    SetBit(FLASH->CR, FLASH_CR_STRT);
    WaitForIdle();
    SetBit(FLASH->SR, FLASH_SR_EOP);
    ClearBit(FLASH->CR, FLASH_CR_PER);
    Lock();
}

//////////////////////////////////////////////////////////////////////////////////////////
//	FLASH_WritePage(): Schreibt den Inhalt von <data> in die Flashspeicherseite Nummer 	//
//	-----------------  <pageNr>. Die betroffene Flashspeicherseite vorher gelöscht.	 	//
//																						//
//  Eingabe: · data:   Zu schreibende Daten.											//
//           · pageNr: Nummer der Ziel-Flashspeicherseite: 0 .. 'FLASH_PAGE_COUNT'-1.	//
//																						//
//  Ausgabe: 'true', wenn die Seite korrekt im Flashspeicher abgelegt wurde.			//
//////////////////////////////////////////////////////////////////////////////////////////
bool
FLASH_WritePage(const UInt08 data[FLASH_PAGE_SIZE], PageNr pageNr)
{
	// (1) Seite löschen
	FLASH_ErasePage(pageNr);
	// (2) Seite beschreiben
	UInt08* flashAddress = FLASH_PageBaseAddress(pageNr);
	Unlock();
	WaitForIdle();
	SetBit(FLASH->CR, FLASH_CR_PG);
	UInt16* flashAddress16 = FLASH_PageBaseAddress(pageNr);
	for (UInt32 i=0; i<FLASH_PAGE_SIZE/2; i++)
	{	WaitForIdle();
		UInt16 data16 = ToWord(data[2*i+1], data[2*i]);
		flashAddress16[i] = data16;
	}
    SetBit(FLASH->SR, FLASH_SR_EOP);
    ClearBit(FLASH->CR, FLASH_CR_PG);
	Lock();
	// (3) Geschriebenes verifizieren
	for (UInt32 i=0; i<FLASH_PAGE_SIZE; i++)
	{   if (flashAddress[i] != data[i]) return false;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
